#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 2
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGEMV(ATL_CINT M0, ATL_CINT N, const TYPE *A, ATL_CINT lda0,
               const TYPE *X, TYPE *Y)
/*
 * 16x2 unrolled mvt_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N2=(N/2)*2, M=M0+M0, M16=((M0/16)*16)<<1,
            lda=lda0+lda0, lda2=lda*2;
   ATL_INT j;

   for (j=N2; j; j -= 2, A += lda2, Y += 4)
   {
      const double *A0=A, *A1=A0+lda;
      register double ry0=ATL_rzero, iy0=ATL_rzero, ry1=ATL_rzero,
                      iy1=ATL_rzero;
      ATL_INT i;
      prefY(Y+4+4-1);
      for (i=0; i < M16; i += 32)
      {
         const register double ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2],
                               ia1_0=A0[i+3], ra2_0=A0[i+4], ia2_0=A0[i+5],
                               ra3_0=A0[i+6], ia3_0=A0[i+7], ra4_0=A0[i+8],
                               ia4_0=A0[i+9], ra5_0=A0[i+10], ia5_0=A0[i+11],
                               ra6_0=A0[i+12], ia6_0=A0[i+13], ra7_0=A0[i+14],
                               ia7_0=A0[i+15], ra8_0=A0[i+16], ia8_0=A0[i+17],
                               ra9_0=A0[i+18], ia9_0=A0[i+19], ra10_0=A0[i+20],
                               ia10_0=A0[i+21], ra11_0=A0[i+22],
                               ia11_0=A0[i+23], ra12_0=A0[i+24],
                               ia12_0=A0[i+25], ra13_0=A0[i+26],
                               ia13_0=A0[i+27], ra14_0=A0[i+28],
                               ia14_0=A0[i+29], ra15_0=A0[i+30],
                               ia15_0=A0[i+31], ra0_1=A1[i+0], ia0_1=A1[i+1],
                               ra1_1=A1[i+2], ia1_1=A1[i+3], ra2_1=A1[i+4],
                               ia2_1=A1[i+5], ra3_1=A1[i+6], ia3_1=A1[i+7],
                               ra4_1=A1[i+8], ia4_1=A1[i+9], ra5_1=A1[i+10],
                               ia5_1=A1[i+11], ra6_1=A1[i+12], ia6_1=A1[i+13],
                               ra7_1=A1[i+14], ia7_1=A1[i+15], ra8_1=A1[i+16],
                               ia8_1=A1[i+17], ra9_1=A1[i+18], ia9_1=A1[i+19],
                               ra10_1=A1[i+20], ia10_1=A1[i+21],
                               ra11_1=A1[i+22], ia11_1=A1[i+23],
                               ra12_1=A1[i+24], ia12_1=A1[i+25],
                               ra13_1=A1[i+26], ia13_1=A1[i+27],
                               ra14_1=A1[i+28], ia14_1=A1[i+29],
                               ra15_1=A1[i+30], ia15_1=A1[i+31];
         const register double rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                               rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                               rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                               rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                               ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                               rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                               ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                               rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                               ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                               rx15=X[i+30], ix15=X[i+31];

         prefX(X);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra0_1 * rx0;
         iy1 += ra0_1 * ix0;
         ry0 += ra1_0 * rx1;
         iy0 += ra1_0 * ix1;
         ry1 += ra1_1 * rx1;
         iy1 += ra1_1 * ix1;
         ry0 += ra2_0 * rx2;
         iy0 += ra2_0 * ix2;
         ry1 += ra2_1 * rx2;
         iy1 += ra2_1 * ix2;
         ry0 += ra3_0 * rx3;
         iy0 += ra3_0 * ix3;
         ry1 += ra3_1 * rx3;
         iy1 += ra3_1 * ix3;
         ry0 += ra4_0 * rx4;
         iy0 += ra4_0 * ix4;
         ry1 += ra4_1 * rx4;
         iy1 += ra4_1 * ix4;
         ry0 += ra5_0 * rx5;
         iy0 += ra5_0 * ix5;
         ry1 += ra5_1 * rx5;
         iy1 += ra5_1 * ix5;
         ry0 += ra6_0 * rx6;
         iy0 += ra6_0 * ix6;
         ry1 += ra6_1 * rx6;
         iy1 += ra6_1 * ix6;
         ry0 += ra7_0 * rx7;
         iy0 += ra7_0 * ix7;
         ry1 += ra7_1 * rx7;
         iy1 += ra7_1 * ix7;
         ry0 += ra8_0 * rx8;
         iy0 += ra8_0 * ix8;
         ry1 += ra8_1 * rx8;
         iy1 += ra8_1 * ix8;
         ry0 += ra9_0 * rx9;
         iy0 += ra9_0 * ix9;
         ry1 += ra9_1 * rx9;
         iy1 += ra9_1 * ix9;
         ry0 += ra10_0 * rx10;
         iy0 += ra10_0 * ix10;
         ry1 += ra10_1 * rx10;
         iy1 += ra10_1 * ix10;
         ry0 += ra11_0 * rx11;
         iy0 += ra11_0 * ix11;
         ry1 += ra11_1 * rx11;
         iy1 += ra11_1 * ix11;
         ry0 += ra12_0 * rx12;
         iy0 += ra12_0 * ix12;
         ry1 += ra12_1 * rx12;
         iy1 += ra12_1 * ix12;
         ry0 += ra13_0 * rx13;
         iy0 += ra13_0 * ix13;
         ry1 += ra13_1 * rx13;
         iy1 += ra13_1 * ix13;
         ry0 += ra14_0 * rx14;
         iy0 += ra14_0 * ix14;
         ry1 += ra14_1 * rx14;
         iy1 += ra14_1 * ix14;
         ry0 += ra15_0 * rx15;
         iy0 += ra15_0 * ix15;
         ry1 += ra15_1 * rx15;
         iy1 += ra15_1 * ix15;

         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry1 -= ia0_1 * ix0;
         iy1 += ia0_1 * rx0;
         prefA(A1);
         ry0 -= ia1_0 * ix1;
         iy0 += ia1_0 * rx1;
         ry1 -= ia1_1 * ix1;
         iy1 += ia1_1 * rx1;
         ry0 -= ia2_0 * ix2;
         iy0 += ia2_0 * rx2;
         ry1 -= ia2_1 * ix2;
         iy1 += ia2_1 * rx2;
         ry0 -= ia3_0 * ix3;
         iy0 += ia3_0 * rx3;
         ry1 -= ia3_1 * ix3;
         iy1 += ia3_1 * rx3;
         ry0 -= ia4_0 * ix4;
         iy0 += ia4_0 * rx4;
         prefA(A0);
         ry1 -= ia4_1 * ix4;
         iy1 += ia4_1 * rx4;
         prefA(A1);
         ry0 -= ia5_0 * ix5;
         iy0 += ia5_0 * rx5;
         ry1 -= ia5_1 * ix5;
         iy1 += ia5_1 * rx5;
         ry0 -= ia6_0 * ix6;
         iy0 += ia6_0 * rx6;
         ry1 -= ia6_1 * ix6;
         iy1 += ia6_1 * rx6;
         ry0 -= ia7_0 * ix7;
         iy0 += ia7_0 * rx7;
         ry1 -= ia7_1 * ix7;
         iy1 += ia7_1 * rx7;
         ry0 -= ia8_0 * ix8;
         iy0 += ia8_0 * rx8;
         prefA(A0);
         ry1 -= ia8_1 * ix8;
         iy1 += ia8_1 * rx8;
         prefA(A1);
         ry0 -= ia9_0 * ix9;
         iy0 += ia9_0 * rx9;
         ry1 -= ia9_1 * ix9;
         iy1 += ia9_1 * rx9;
         ry0 -= ia10_0 * ix10;
         iy0 += ia10_0 * rx10;
         ry1 -= ia10_1 * ix10;
         iy1 += ia10_1 * rx10;
         ry0 -= ia11_0 * ix11;
         iy0 += ia11_0 * rx11;
         ry1 -= ia11_1 * ix11;
         iy1 += ia11_1 * rx11;
         ry0 -= ia12_0 * ix12;
         iy0 += ia12_0 * rx12;
         prefA(A0);
         ry1 -= ia12_1 * ix12;
         iy1 += ia12_1 * rx12;
         prefA(A1);
         ry0 -= ia13_0 * ix13;
         iy0 += ia13_0 * rx13;
         ry1 -= ia13_1 * ix13;
         iy1 += ia13_1 * rx13;
         ry0 -= ia14_0 * ix14;
         iy0 += ia14_0 * rx14;
         ry1 -= ia14_1 * ix14;
         iy1 += ia14_1 * rx14;
         ry0 -= ia15_0 * ix15;
         iy0 += ia15_0 * rx15;
         ry1 -= ia15_1 * ix15;
         iy1 += ia15_1 * rx15;
      }
      for (i=M16; i < M; i += 2)
      {
         const register double rx0 = X[i], ix0 = X[i+1];
         const register double ra0_0=A0[i], ia0_0=A0[i+1], ra0_1=A1[i],
                               ia0_1=A1[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra0_1 * rx0;
         iy1 += ra0_1 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         ry1 -= ia0_1 * ix0;
         iy1 += ia0_1 * rx0;
      }
      #ifdef BETA0
         Y[0] = ry0;
         Y[1] = iy0;
         Y[2] = ry1;
         Y[3] = iy1;
      #else
         Y[0] += ry0;
         Y[1] += iy0;
         Y[2] += ry1;
         Y[3] += iy1;
      #endif
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N2; j; j--, A += lda, Y += 2)
   {
      const double *A0=A;
      register double ry0=ATL_rzero, iy0=ATL_rzero;
      ATL_INT i;
      prefY(Y+4+4-1);
      for (i=0; i < M16; i += 32)
      {
         const register double ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2],
                               ia1_0=A0[i+3], ra2_0=A0[i+4], ia2_0=A0[i+5],
                               ra3_0=A0[i+6], ia3_0=A0[i+7], ra4_0=A0[i+8],
                               ia4_0=A0[i+9], ra5_0=A0[i+10], ia5_0=A0[i+11],
                               ra6_0=A0[i+12], ia6_0=A0[i+13], ra7_0=A0[i+14],
                               ia7_0=A0[i+15], ra8_0=A0[i+16], ia8_0=A0[i+17],
                               ra9_0=A0[i+18], ia9_0=A0[i+19], ra10_0=A0[i+20],
                               ia10_0=A0[i+21], ra11_0=A0[i+22],
                               ia11_0=A0[i+23], ra12_0=A0[i+24],
                               ia12_0=A0[i+25], ra13_0=A0[i+26],
                               ia13_0=A0[i+27], ra14_0=A0[i+28],
                               ia14_0=A0[i+29], ra15_0=A0[i+30],
                               ia15_0=A0[i+31];
         const register double rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                               rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                               rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                               rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                               ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                               rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                               ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                               rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                               ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                               rx15=X[i+30], ix15=X[i+31];

         prefX(X);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 += ra1_0 * rx1;
         iy0 += ra1_0 * ix1;
         ry0 += ra2_0 * rx2;
         iy0 += ra2_0 * ix2;
         ry0 += ra3_0 * rx3;
         iy0 += ra3_0 * ix3;
         ry0 += ra4_0 * rx4;
         iy0 += ra4_0 * ix4;
         ry0 += ra5_0 * rx5;
         iy0 += ra5_0 * ix5;
         ry0 += ra6_0 * rx6;
         iy0 += ra6_0 * ix6;
         ry0 += ra7_0 * rx7;
         iy0 += ra7_0 * ix7;
         ry0 += ra8_0 * rx8;
         iy0 += ra8_0 * ix8;
         ry0 += ra9_0 * rx9;
         iy0 += ra9_0 * ix9;
         ry0 += ra10_0 * rx10;
         iy0 += ra10_0 * ix10;
         ry0 += ra11_0 * rx11;
         iy0 += ra11_0 * ix11;
         ry0 += ra12_0 * rx12;
         iy0 += ra12_0 * ix12;
         ry0 += ra13_0 * rx13;
         iy0 += ra13_0 * ix13;
         ry0 += ra14_0 * rx14;
         iy0 += ra14_0 * ix14;
         ry0 += ra15_0 * rx15;
         iy0 += ra15_0 * ix15;

         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry0 -= ia1_0 * ix1;
         iy0 += ia1_0 * rx1;
         ry0 -= ia2_0 * ix2;
         iy0 += ia2_0 * rx2;
         ry0 -= ia3_0 * ix3;
         iy0 += ia3_0 * rx3;
         ry0 -= ia4_0 * ix4;
         iy0 += ia4_0 * rx4;
         prefA(A0);
         ry0 -= ia5_0 * ix5;
         iy0 += ia5_0 * rx5;
         ry0 -= ia6_0 * ix6;
         iy0 += ia6_0 * rx6;
         ry0 -= ia7_0 * ix7;
         iy0 += ia7_0 * rx7;
         ry0 -= ia8_0 * ix8;
         iy0 += ia8_0 * rx8;
         prefA(A0);
         ry0 -= ia9_0 * ix9;
         iy0 += ia9_0 * rx9;
         ry0 -= ia10_0 * ix10;
         iy0 += ia10_0 * rx10;
         ry0 -= ia11_0 * ix11;
         iy0 += ia11_0 * rx11;
         ry0 -= ia12_0 * ix12;
         iy0 += ia12_0 * rx12;
         prefA(A0);
         ry0 -= ia13_0 * ix13;
         iy0 += ia13_0 * rx13;
         ry0 -= ia14_0 * ix14;
         iy0 += ia14_0 * rx14;
         ry0 -= ia15_0 * ix15;
         iy0 += ia15_0 * rx15;
      }
      for (i=M16; i < M; i += 2)
      {
         const register double rx0 = X[i], ix0 = X[i+1];
         const register double ra0_0=A0[i], ia0_0=A0[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
      }
      #ifdef BETA0
         Y[0] = ry0;
         Y[1] = iy0;
      #else
         Y[0] += ry0;
         Y[1] += iy0;
      #endif
   }
}
