#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 8
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGERK(ATL_CINT M0, ATL_CINT N, const double *X, const double *Y, 
               double *A, ATL_CINT lda0)
/*
 * 1x8 unrolled r1_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N8=(N/8)*8, M=M0+M0, M1=((M0/1)*1)<<1,
            lda=lda0+lda0, lda8=lda*8;
   ATL_INT j;

   for (j=N8; j; j -= 8, A += lda8, Y += 16)
   {
      double *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda, *A5=A4+lda,
             *A6=A5+lda, *A7=A6+lda;
      const register double ry0=Y[0], iy0=Y[1], ry1=Y[2], iy1=Y[3], ry2=Y[4],
                            iy2=Y[5], ry3=Y[6], iy3=Y[7], ry4=Y[8], iy4=Y[9],
                            ry5=Y[10], iy5=Y[11], ry6=Y[12], iy6=Y[13],
                            ry7=Y[14], iy7=Y[15];
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M1; i += 2)
      {
         const register double rx0=X[i+0], ix0=X[i+1];

         prefX(X);
         A0[i+0] += ry0*rx0 - iy0*ix0;
         prefA(A0);
         A0[i+0+1] += ry0*ix0 + iy0*rx0;
         A1[i+0] += ry1*rx0 - iy1*ix0;
         prefA(A1);
         A1[i+0+1] += ry1*ix0 + iy1*rx0;
         A2[i+0] += ry2*rx0 - iy2*ix0;
         prefA(A2);
         A2[i+0+1] += ry2*ix0 + iy2*rx0;
         A3[i+0] += ry3*rx0 - iy3*ix0;
         prefA(A3);
         A3[i+0+1] += ry3*ix0 + iy3*rx0;
         A4[i+0] += ry4*rx0 - iy4*ix0;
         prefA(A4);
         A4[i+0+1] += ry4*ix0 + iy4*rx0;
         A5[i+0] += ry5*rx0 - iy5*ix0;
         prefA(A5);
         A5[i+0+1] += ry5*ix0 + iy5*rx0;
         A6[i+0] += ry6*rx0 - iy6*ix0;
         prefA(A6);
         A6[i+0+1] += ry6*ix0 + iy6*rx0;
         A7[i+0] += ry7*rx0 - iy7*ix0;
         prefA(A7);
         A7[i+0+1] += ry7*ix0 + iy7*rx0;
      }
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N8; j; j--, A += lda, Y += 2)
   {
      double *A0=A;
      const register double ry0=Y[0], iy0=Y[1];
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M1; i += 2)
      {
         const register double rx0=X[i+0], ix0=X[i+1];

         prefX(X);
         A0[i+0] += ry0*rx0 - iy0*ix0;
         prefA(A0);
         A0[i+0+1] += ry0*ix0 + iy0*rx0;
      }
   }
}
