#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 12
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
/* Need to handle BETA=0 case by assigning y to zero outside of loop */
void ATL_UGEMV(ATL_CINT M0, ATL_CINT N, const TYPE *A, ATL_CINT lda0,
               const TYPE *X, TYPE *Y)
/*
 * 8x12 unrolled mvn_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N12=(N/12)*12, M=M0+M0, M8=((M0/8)*8)<<1,
            lda=lda0+lda0, lda12=lda*12;
   ATL_INT j;

   #ifdef BETA0
      for (j=0; j < M; j++)
         Y[j] = ATL_rzero;
   #endif
   for (j=N12; j; j -= 12, A += lda12, X += 24)
   {
      const double *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda,
                   *A5=A4+lda, *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda,
                   *A10=A9+lda, *A11=A10+lda;
      register double rx0=X[0], ix0=X[1], rx1=X[2], ix1=X[3], rx2=X[4],
                      ix2=X[5], rx3=X[6], ix3=X[7], rx4=X[8], ix4=X[9],
                      rx5=X[10], ix5=X[11], rx6=X[12], ix6=X[13], rx7=X[14],
                      ix7=X[15], rx8=X[16], ix8=X[17], rx9=X[18], ix9=X[19],
                      rx10=X[20], ix10=X[21], rx11=X[22], ix11=X[23];
      ATL_INT i;
      prefY(X+24+24-1);
      for (i=0; i < M8; i += 16)
      {
         const double ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2],
                      ia1_0=A0[i+3], ra2_0=A0[i+4], ia2_0=A0[i+5],
                      ra3_0=A0[i+6], ia3_0=A0[i+7], ra4_0=A0[i+8],
                      ia4_0=A0[i+9], ra5_0=A0[i+10], ia5_0=A0[i+11],
                      ra6_0=A0[i+12], ia6_0=A0[i+13], ra7_0=A0[i+14],
                      ia7_0=A0[i+15], ra0_1=A1[i+0], ia0_1=A1[i+1],
                      ra1_1=A1[i+2], ia1_1=A1[i+3], ra2_1=A1[i+4],
                      ia2_1=A1[i+5], ra3_1=A1[i+6], ia3_1=A1[i+7],
                      ra4_1=A1[i+8], ia4_1=A1[i+9], ra5_1=A1[i+10],
                      ia5_1=A1[i+11], ra6_1=A1[i+12], ia6_1=A1[i+13],
                      ra7_1=A1[i+14], ia7_1=A1[i+15], ra0_2=A2[i+0],
                      ia0_2=A2[i+1], ra1_2=A2[i+2], ia1_2=A2[i+3],
                      ra2_2=A2[i+4], ia2_2=A2[i+5], ra3_2=A2[i+6],
                      ia3_2=A2[i+7], ra4_2=A2[i+8], ia4_2=A2[i+9],
                      ra5_2=A2[i+10], ia5_2=A2[i+11], ra6_2=A2[i+12],
                      ia6_2=A2[i+13], ra7_2=A2[i+14], ia7_2=A2[i+15],
                      ra0_3=A3[i+0], ia0_3=A3[i+1], ra1_3=A3[i+2],
                      ia1_3=A3[i+3], ra2_3=A3[i+4], ia2_3=A3[i+5],
                      ra3_3=A3[i+6], ia3_3=A3[i+7], ra4_3=A3[i+8],
                      ia4_3=A3[i+9], ra5_3=A3[i+10], ia5_3=A3[i+11],
                      ra6_3=A3[i+12], ia6_3=A3[i+13], ra7_3=A3[i+14],
                      ia7_3=A3[i+15], ra0_4=A4[i+0], ia0_4=A4[i+1],
                      ra1_4=A4[i+2], ia1_4=A4[i+3], ra2_4=A4[i+4],
                      ia2_4=A4[i+5], ra3_4=A4[i+6], ia3_4=A4[i+7],
                      ra4_4=A4[i+8], ia4_4=A4[i+9], ra5_4=A4[i+10],
                      ia5_4=A4[i+11], ra6_4=A4[i+12], ia6_4=A4[i+13],
                      ra7_4=A4[i+14], ia7_4=A4[i+15], ra0_5=A5[i+0],
                      ia0_5=A5[i+1], ra1_5=A5[i+2], ia1_5=A5[i+3],
                      ra2_5=A5[i+4], ia2_5=A5[i+5], ra3_5=A5[i+6],
                      ia3_5=A5[i+7], ra4_5=A5[i+8], ia4_5=A5[i+9],
                      ra5_5=A5[i+10], ia5_5=A5[i+11], ra6_5=A5[i+12],
                      ia6_5=A5[i+13], ra7_5=A5[i+14], ia7_5=A5[i+15],
                      ra0_6=A6[i+0], ia0_6=A6[i+1], ra1_6=A6[i+2],
                      ia1_6=A6[i+3], ra2_6=A6[i+4], ia2_6=A6[i+5],
                      ra3_6=A6[i+6], ia3_6=A6[i+7], ra4_6=A6[i+8],
                      ia4_6=A6[i+9], ra5_6=A6[i+10], ia5_6=A6[i+11],
                      ra6_6=A6[i+12], ia6_6=A6[i+13], ra7_6=A6[i+14],
                      ia7_6=A6[i+15], ra0_7=A7[i+0], ia0_7=A7[i+1],
                      ra1_7=A7[i+2], ia1_7=A7[i+3], ra2_7=A7[i+4],
                      ia2_7=A7[i+5], ra3_7=A7[i+6], ia3_7=A7[i+7],
                      ra4_7=A7[i+8], ia4_7=A7[i+9], ra5_7=A7[i+10],
                      ia5_7=A7[i+11], ra6_7=A7[i+12], ia6_7=A7[i+13],
                      ra7_7=A7[i+14], ia7_7=A7[i+15], ra0_8=A8[i+0],
                      ia0_8=A8[i+1], ra1_8=A8[i+2], ia1_8=A8[i+3],
                      ra2_8=A8[i+4], ia2_8=A8[i+5], ra3_8=A8[i+6],
                      ia3_8=A8[i+7], ra4_8=A8[i+8], ia4_8=A8[i+9],
                      ra5_8=A8[i+10], ia5_8=A8[i+11], ra6_8=A8[i+12],
                      ia6_8=A8[i+13], ra7_8=A8[i+14], ia7_8=A8[i+15],
                      ra0_9=A9[i+0], ia0_9=A9[i+1], ra1_9=A9[i+2],
                      ia1_9=A9[i+3], ra2_9=A9[i+4], ia2_9=A9[i+5],
                      ra3_9=A9[i+6], ia3_9=A9[i+7], ra4_9=A9[i+8],
                      ia4_9=A9[i+9], ra5_9=A9[i+10], ia5_9=A9[i+11],
                      ra6_9=A9[i+12], ia6_9=A9[i+13], ra7_9=A9[i+14],
                      ia7_9=A9[i+15], ra0_10=A10[i+0], ia0_10=A10[i+1],
                      ra1_10=A10[i+2], ia1_10=A10[i+3], ra2_10=A10[i+4],
                      ia2_10=A10[i+5], ra3_10=A10[i+6], ia3_10=A10[i+7],
                      ra4_10=A10[i+8], ia4_10=A10[i+9], ra5_10=A10[i+10],
                      ia5_10=A10[i+11], ra6_10=A10[i+12], ia6_10=A10[i+13],
                      ra7_10=A10[i+14], ia7_10=A10[i+15], ra0_11=A11[i+0],
                      ia0_11=A11[i+1], ra1_11=A11[i+2], ia1_11=A11[i+3],
                      ra2_11=A11[i+4], ia2_11=A11[i+5], ra3_11=A11[i+6],
                      ia3_11=A11[i+7], ra4_11=A11[i+8], ia4_11=A11[i+9],
                      ra5_11=A11[i+10], ia5_11=A11[i+11], ra6_11=A11[i+12],
                      ia6_11=A11[i+13], ra7_11=A11[i+14], ia7_11=A11[i+15];
         register double ry0=Y[i+0], iy0=Y[i+1], ry1=Y[i+2], iy1=Y[i+3],
                         ry2=Y[i+4], iy2=Y[i+5], ry3=Y[i+6], iy3=Y[i+7],
                         ry4=Y[i+8], iy4=Y[i+9], ry5=Y[i+10], iy5=Y[i+11],
                         ry6=Y[i+12], iy6=Y[i+13], ry7=Y[i+14], iy7=Y[i+15];

         prefY(Y);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra1_0 * rx0;
         iy1 += ra1_0 * ix0;
         ry2 += ra2_0 * rx0;
         iy2 += ra2_0 * ix0;
         ry3 += ra3_0 * rx0;
         iy3 += ra3_0 * ix0;
         ry4 += ra4_0 * rx0;
         iy4 += ra4_0 * ix0;
         ry5 += ra5_0 * rx0;
         iy5 += ra5_0 * ix0;
         ry6 += ra6_0 * rx0;
         iy6 += ra6_0 * ix0;
         ry7 += ra7_0 * rx0;
         iy7 += ra7_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry1 -= ia1_0 * ix0;
         iy1 += ia1_0 * rx0;
         ry2 -= ia2_0 * ix0;
         iy2 += ia2_0 * rx0;
         ry3 -= ia3_0 * ix0;
         iy3 += ia3_0 * rx0;
         ry4 -= ia4_0 * ix0;
         iy4 += ia4_0 * rx0;
         prefA(A0);
         ry5 -= ia5_0 * ix0;
         iy5 += ia5_0 * rx0;
         ry6 -= ia6_0 * ix0;
         iy6 += ia6_0 * rx0;
         ry7 -= ia7_0 * ix0;
         iy7 += ia7_0 * rx0;

         ry0 += ra0_1 * rx1;
         iy0 += ra0_1 * ix1;
         ry1 += ra1_1 * rx1;
         iy1 += ra1_1 * ix1;
         ry2 += ra2_1 * rx1;
         iy2 += ra2_1 * ix1;
         ry3 += ra3_1 * rx1;
         iy3 += ra3_1 * ix1;
         ry4 += ra4_1 * rx1;
         iy4 += ra4_1 * ix1;
         ry5 += ra5_1 * rx1;
         iy5 += ra5_1 * ix1;
         ry6 += ra6_1 * rx1;
         iy6 += ra6_1 * ix1;
         ry7 += ra7_1 * rx1;
         iy7 += ra7_1 * ix1;
         ry0 -= ia0_1 * ix1;
         iy0 += ia0_1 * rx1;
         prefA(A1);
         ry1 -= ia1_1 * ix1;
         iy1 += ia1_1 * rx1;
         ry2 -= ia2_1 * ix1;
         iy2 += ia2_1 * rx1;
         ry3 -= ia3_1 * ix1;
         iy3 += ia3_1 * rx1;
         ry4 -= ia4_1 * ix1;
         iy4 += ia4_1 * rx1;
         prefA(A1);
         ry5 -= ia5_1 * ix1;
         iy5 += ia5_1 * rx1;
         ry6 -= ia6_1 * ix1;
         iy6 += ia6_1 * rx1;
         ry7 -= ia7_1 * ix1;
         iy7 += ia7_1 * rx1;

         ry0 += ra0_2 * rx2;
         iy0 += ra0_2 * ix2;
         ry1 += ra1_2 * rx2;
         iy1 += ra1_2 * ix2;
         ry2 += ra2_2 * rx2;
         iy2 += ra2_2 * ix2;
         ry3 += ra3_2 * rx2;
         iy3 += ra3_2 * ix2;
         ry4 += ra4_2 * rx2;
         iy4 += ra4_2 * ix2;
         ry5 += ra5_2 * rx2;
         iy5 += ra5_2 * ix2;
         ry6 += ra6_2 * rx2;
         iy6 += ra6_2 * ix2;
         ry7 += ra7_2 * rx2;
         iy7 += ra7_2 * ix2;
         ry0 -= ia0_2 * ix2;
         iy0 += ia0_2 * rx2;
         prefA(A2);
         ry1 -= ia1_2 * ix2;
         iy1 += ia1_2 * rx2;
         ry2 -= ia2_2 * ix2;
         iy2 += ia2_2 * rx2;
         ry3 -= ia3_2 * ix2;
         iy3 += ia3_2 * rx2;
         ry4 -= ia4_2 * ix2;
         iy4 += ia4_2 * rx2;
         prefA(A2);
         ry5 -= ia5_2 * ix2;
         iy5 += ia5_2 * rx2;
         ry6 -= ia6_2 * ix2;
         iy6 += ia6_2 * rx2;
         ry7 -= ia7_2 * ix2;
         iy7 += ia7_2 * rx2;

         ry0 += ra0_3 * rx3;
         iy0 += ra0_3 * ix3;
         ry1 += ra1_3 * rx3;
         iy1 += ra1_3 * ix3;
         ry2 += ra2_3 * rx3;
         iy2 += ra2_3 * ix3;
         ry3 += ra3_3 * rx3;
         iy3 += ra3_3 * ix3;
         ry4 += ra4_3 * rx3;
         iy4 += ra4_3 * ix3;
         ry5 += ra5_3 * rx3;
         iy5 += ra5_3 * ix3;
         ry6 += ra6_3 * rx3;
         iy6 += ra6_3 * ix3;
         ry7 += ra7_3 * rx3;
         iy7 += ra7_3 * ix3;
         ry0 -= ia0_3 * ix3;
         iy0 += ia0_3 * rx3;
         prefA(A3);
         ry1 -= ia1_3 * ix3;
         iy1 += ia1_3 * rx3;
         ry2 -= ia2_3 * ix3;
         iy2 += ia2_3 * rx3;
         ry3 -= ia3_3 * ix3;
         iy3 += ia3_3 * rx3;
         ry4 -= ia4_3 * ix3;
         iy4 += ia4_3 * rx3;
         prefA(A3);
         ry5 -= ia5_3 * ix3;
         iy5 += ia5_3 * rx3;
         ry6 -= ia6_3 * ix3;
         iy6 += ia6_3 * rx3;
         ry7 -= ia7_3 * ix3;
         iy7 += ia7_3 * rx3;

         ry0 += ra0_4 * rx4;
         iy0 += ra0_4 * ix4;
         ry1 += ra1_4 * rx4;
         iy1 += ra1_4 * ix4;
         ry2 += ra2_4 * rx4;
         iy2 += ra2_4 * ix4;
         ry3 += ra3_4 * rx4;
         iy3 += ra3_4 * ix4;
         ry4 += ra4_4 * rx4;
         iy4 += ra4_4 * ix4;
         ry5 += ra5_4 * rx4;
         iy5 += ra5_4 * ix4;
         ry6 += ra6_4 * rx4;
         iy6 += ra6_4 * ix4;
         ry7 += ra7_4 * rx4;
         iy7 += ra7_4 * ix4;
         ry0 -= ia0_4 * ix4;
         iy0 += ia0_4 * rx4;
         prefA(A4);
         ry1 -= ia1_4 * ix4;
         iy1 += ia1_4 * rx4;
         ry2 -= ia2_4 * ix4;
         iy2 += ia2_4 * rx4;
         ry3 -= ia3_4 * ix4;
         iy3 += ia3_4 * rx4;
         ry4 -= ia4_4 * ix4;
         iy4 += ia4_4 * rx4;
         prefA(A4);
         ry5 -= ia5_4 * ix4;
         iy5 += ia5_4 * rx4;
         ry6 -= ia6_4 * ix4;
         iy6 += ia6_4 * rx4;
         ry7 -= ia7_4 * ix4;
         iy7 += ia7_4 * rx4;

         ry0 += ra0_5 * rx5;
         iy0 += ra0_5 * ix5;
         ry1 += ra1_5 * rx5;
         iy1 += ra1_5 * ix5;
         ry2 += ra2_5 * rx5;
         iy2 += ra2_5 * ix5;
         ry3 += ra3_5 * rx5;
         iy3 += ra3_5 * ix5;
         ry4 += ra4_5 * rx5;
         iy4 += ra4_5 * ix5;
         ry5 += ra5_5 * rx5;
         iy5 += ra5_5 * ix5;
         ry6 += ra6_5 * rx5;
         iy6 += ra6_5 * ix5;
         ry7 += ra7_5 * rx5;
         iy7 += ra7_5 * ix5;
         ry0 -= ia0_5 * ix5;
         iy0 += ia0_5 * rx5;
         prefA(A5);
         ry1 -= ia1_5 * ix5;
         iy1 += ia1_5 * rx5;
         ry2 -= ia2_5 * ix5;
         iy2 += ia2_5 * rx5;
         ry3 -= ia3_5 * ix5;
         iy3 += ia3_5 * rx5;
         ry4 -= ia4_5 * ix5;
         iy4 += ia4_5 * rx5;
         prefA(A5);
         ry5 -= ia5_5 * ix5;
         iy5 += ia5_5 * rx5;
         ry6 -= ia6_5 * ix5;
         iy6 += ia6_5 * rx5;
         ry7 -= ia7_5 * ix5;
         iy7 += ia7_5 * rx5;

         ry0 += ra0_6 * rx6;
         iy0 += ra0_6 * ix6;
         ry1 += ra1_6 * rx6;
         iy1 += ra1_6 * ix6;
         ry2 += ra2_6 * rx6;
         iy2 += ra2_6 * ix6;
         ry3 += ra3_6 * rx6;
         iy3 += ra3_6 * ix6;
         ry4 += ra4_6 * rx6;
         iy4 += ra4_6 * ix6;
         ry5 += ra5_6 * rx6;
         iy5 += ra5_6 * ix6;
         ry6 += ra6_6 * rx6;
         iy6 += ra6_6 * ix6;
         ry7 += ra7_6 * rx6;
         iy7 += ra7_6 * ix6;
         ry0 -= ia0_6 * ix6;
         iy0 += ia0_6 * rx6;
         prefA(A6);
         ry1 -= ia1_6 * ix6;
         iy1 += ia1_6 * rx6;
         ry2 -= ia2_6 * ix6;
         iy2 += ia2_6 * rx6;
         ry3 -= ia3_6 * ix6;
         iy3 += ia3_6 * rx6;
         ry4 -= ia4_6 * ix6;
         iy4 += ia4_6 * rx6;
         prefA(A6);
         ry5 -= ia5_6 * ix6;
         iy5 += ia5_6 * rx6;
         ry6 -= ia6_6 * ix6;
         iy6 += ia6_6 * rx6;
         ry7 -= ia7_6 * ix6;
         iy7 += ia7_6 * rx6;

         ry0 += ra0_7 * rx7;
         iy0 += ra0_7 * ix7;
         ry1 += ra1_7 * rx7;
         iy1 += ra1_7 * ix7;
         ry2 += ra2_7 * rx7;
         iy2 += ra2_7 * ix7;
         ry3 += ra3_7 * rx7;
         iy3 += ra3_7 * ix7;
         ry4 += ra4_7 * rx7;
         iy4 += ra4_7 * ix7;
         ry5 += ra5_7 * rx7;
         iy5 += ra5_7 * ix7;
         ry6 += ra6_7 * rx7;
         iy6 += ra6_7 * ix7;
         ry7 += ra7_7 * rx7;
         iy7 += ra7_7 * ix7;
         ry0 -= ia0_7 * ix7;
         iy0 += ia0_7 * rx7;
         prefA(A7);
         ry1 -= ia1_7 * ix7;
         iy1 += ia1_7 * rx7;
         ry2 -= ia2_7 * ix7;
         iy2 += ia2_7 * rx7;
         ry3 -= ia3_7 * ix7;
         iy3 += ia3_7 * rx7;
         ry4 -= ia4_7 * ix7;
         iy4 += ia4_7 * rx7;
         prefA(A7);
         ry5 -= ia5_7 * ix7;
         iy5 += ia5_7 * rx7;
         ry6 -= ia6_7 * ix7;
         iy6 += ia6_7 * rx7;
         ry7 -= ia7_7 * ix7;
         iy7 += ia7_7 * rx7;

         ry0 += ra0_8 * rx8;
         iy0 += ra0_8 * ix8;
         ry1 += ra1_8 * rx8;
         iy1 += ra1_8 * ix8;
         ry2 += ra2_8 * rx8;
         iy2 += ra2_8 * ix8;
         ry3 += ra3_8 * rx8;
         iy3 += ra3_8 * ix8;
         ry4 += ra4_8 * rx8;
         iy4 += ra4_8 * ix8;
         ry5 += ra5_8 * rx8;
         iy5 += ra5_8 * ix8;
         ry6 += ra6_8 * rx8;
         iy6 += ra6_8 * ix8;
         ry7 += ra7_8 * rx8;
         iy7 += ra7_8 * ix8;
         ry0 -= ia0_8 * ix8;
         iy0 += ia0_8 * rx8;
         prefA(A8);
         ry1 -= ia1_8 * ix8;
         iy1 += ia1_8 * rx8;
         ry2 -= ia2_8 * ix8;
         iy2 += ia2_8 * rx8;
         ry3 -= ia3_8 * ix8;
         iy3 += ia3_8 * rx8;
         ry4 -= ia4_8 * ix8;
         iy4 += ia4_8 * rx8;
         prefA(A8);
         ry5 -= ia5_8 * ix8;
         iy5 += ia5_8 * rx8;
         ry6 -= ia6_8 * ix8;
         iy6 += ia6_8 * rx8;
         ry7 -= ia7_8 * ix8;
         iy7 += ia7_8 * rx8;

         ry0 += ra0_9 * rx9;
         iy0 += ra0_9 * ix9;
         ry1 += ra1_9 * rx9;
         iy1 += ra1_9 * ix9;
         ry2 += ra2_9 * rx9;
         iy2 += ra2_9 * ix9;
         ry3 += ra3_9 * rx9;
         iy3 += ra3_9 * ix9;
         ry4 += ra4_9 * rx9;
         iy4 += ra4_9 * ix9;
         ry5 += ra5_9 * rx9;
         iy5 += ra5_9 * ix9;
         ry6 += ra6_9 * rx9;
         iy6 += ra6_9 * ix9;
         ry7 += ra7_9 * rx9;
         iy7 += ra7_9 * ix9;
         ry0 -= ia0_9 * ix9;
         iy0 += ia0_9 * rx9;
         prefA(A9);
         ry1 -= ia1_9 * ix9;
         iy1 += ia1_9 * rx9;
         ry2 -= ia2_9 * ix9;
         iy2 += ia2_9 * rx9;
         ry3 -= ia3_9 * ix9;
         iy3 += ia3_9 * rx9;
         ry4 -= ia4_9 * ix9;
         iy4 += ia4_9 * rx9;
         prefA(A9);
         ry5 -= ia5_9 * ix9;
         iy5 += ia5_9 * rx9;
         ry6 -= ia6_9 * ix9;
         iy6 += ia6_9 * rx9;
         ry7 -= ia7_9 * ix9;
         iy7 += ia7_9 * rx9;

         ry0 += ra0_10 * rx10;
         iy0 += ra0_10 * ix10;
         ry1 += ra1_10 * rx10;
         iy1 += ra1_10 * ix10;
         ry2 += ra2_10 * rx10;
         iy2 += ra2_10 * ix10;
         ry3 += ra3_10 * rx10;
         iy3 += ra3_10 * ix10;
         ry4 += ra4_10 * rx10;
         iy4 += ra4_10 * ix10;
         ry5 += ra5_10 * rx10;
         iy5 += ra5_10 * ix10;
         ry6 += ra6_10 * rx10;
         iy6 += ra6_10 * ix10;
         ry7 += ra7_10 * rx10;
         iy7 += ra7_10 * ix10;
         ry0 -= ia0_10 * ix10;
         iy0 += ia0_10 * rx10;
         prefA(A10);
         ry1 -= ia1_10 * ix10;
         iy1 += ia1_10 * rx10;
         ry2 -= ia2_10 * ix10;
         iy2 += ia2_10 * rx10;
         ry3 -= ia3_10 * ix10;
         iy3 += ia3_10 * rx10;
         ry4 -= ia4_10 * ix10;
         iy4 += ia4_10 * rx10;
         prefA(A10);
         ry5 -= ia5_10 * ix10;
         iy5 += ia5_10 * rx10;
         ry6 -= ia6_10 * ix10;
         iy6 += ia6_10 * rx10;
         ry7 -= ia7_10 * ix10;
         iy7 += ia7_10 * rx10;

         ry0 += ra0_11 * rx11;
         iy0 += ra0_11 * ix11;
         ry1 += ra1_11 * rx11;
         iy1 += ra1_11 * ix11;
         ry2 += ra2_11 * rx11;
         iy2 += ra2_11 * ix11;
         ry3 += ra3_11 * rx11;
         iy3 += ra3_11 * ix11;
         ry4 += ra4_11 * rx11;
         iy4 += ra4_11 * ix11;
         ry5 += ra5_11 * rx11;
         iy5 += ra5_11 * ix11;
         ry6 += ra6_11 * rx11;
         iy6 += ra6_11 * ix11;
         ry7 += ra7_11 * rx11;
         iy7 += ra7_11 * ix11;
         ry0 -= ia0_11 * ix11;
         iy0 += ia0_11 * rx11;
         prefA(A11);
         ry1 -= ia1_11 * ix11;
         iy1 += ia1_11 * rx11;
         ry2 -= ia2_11 * ix11;
         iy2 += ia2_11 * rx11;
         ry3 -= ia3_11 * ix11;
         iy3 += ia3_11 * rx11;
         ry4 -= ia4_11 * ix11;
         iy4 += ia4_11 * rx11;
         prefA(A11);
         ry5 -= ia5_11 * ix11;
         iy5 += ia5_11 * rx11;
         ry6 -= ia6_11 * ix11;
         iy6 += ia6_11 * rx11;
         ry7 -= ia7_11 * ix11;
         iy7 += ia7_11 * rx11;


         Y[i+0] = ry0;
         Y[i+1] = iy0;
         Y[i+2] = ry1;
         Y[i+3] = iy1;
         Y[i+4] = ry2;
         Y[i+5] = iy2;
         Y[i+6] = ry3;
         Y[i+7] = iy3;
         Y[i+8] = ry4;
         Y[i+9] = iy4;
         Y[i+10] = ry5;
         Y[i+11] = iy5;
         Y[i+12] = ry6;
         Y[i+13] = iy6;
         Y[i+14] = ry7;
         Y[i+15] = iy7;
      }
      for (i=M8; i < M; i += 2)
      {
         register double ry0 = Y[i], iy0 = Y[i+1];
         const double ra0_0=A0[i], ia0_0=A0[i+1], ra0_1=A1[i], ia0_1=A1[i+1],
                      ra0_2=A2[i], ia0_2=A2[i+1], ra0_3=A3[i], ia0_3=A3[i+1],
                      ra0_4=A4[i], ia0_4=A4[i+1], ra0_5=A5[i], ia0_5=A5[i+1],
                      ra0_6=A6[i], ia0_6=A6[i+1], ra0_7=A7[i], ia0_7=A7[i+1],
                      ra0_8=A8[i], ia0_8=A8[i+1], ra0_9=A9[i], ia0_9=A9[i+1],
                      ra0_10=A10[i], ia0_10=A10[i+1], ra0_11=A11[i],
                      ia0_11=A11[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         ry0 += ra0_1 * rx1;
         iy0 += ra0_1 * ix1;
         ry0 -= ia0_1 * ix1;
         iy0 += ia0_1 * rx1;
         ry0 += ra0_2 * rx2;
         iy0 += ra0_2 * ix2;
         ry0 -= ia0_2 * ix2;
         iy0 += ia0_2 * rx2;
         ry0 += ra0_3 * rx3;
         iy0 += ra0_3 * ix3;
         ry0 -= ia0_3 * ix3;
         iy0 += ia0_3 * rx3;
         ry0 += ra0_4 * rx4;
         iy0 += ra0_4 * ix4;
         ry0 -= ia0_4 * ix4;
         iy0 += ia0_4 * rx4;
         ry0 += ra0_5 * rx5;
         iy0 += ra0_5 * ix5;
         ry0 -= ia0_5 * ix5;
         iy0 += ia0_5 * rx5;
         ry0 += ra0_6 * rx6;
         iy0 += ra0_6 * ix6;
         ry0 -= ia0_6 * ix6;
         iy0 += ia0_6 * rx6;
         ry0 += ra0_7 * rx7;
         iy0 += ra0_7 * ix7;
         ry0 -= ia0_7 * ix7;
         iy0 += ia0_7 * rx7;
         ry0 += ra0_8 * rx8;
         iy0 += ra0_8 * ix8;
         ry0 -= ia0_8 * ix8;
         iy0 += ia0_8 * rx8;
         ry0 += ra0_9 * rx9;
         iy0 += ra0_9 * ix9;
         ry0 -= ia0_9 * ix9;
         iy0 += ia0_9 * rx9;
         ry0 += ra0_10 * rx10;
         iy0 += ra0_10 * ix10;
         ry0 -= ia0_10 * ix10;
         iy0 += ia0_10 * rx10;
         ry0 += ra0_11 * rx11;
         iy0 += ra0_11 * ix11;
         ry0 -= ia0_11 * ix11;
         iy0 += ia0_11 * rx11;
         Y[i] = ry0;
         Y[i+1] = iy0;
      }
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N12; j; j--, A += lda, X += 2)
   {
      const double *A0=A;
      register double rx0=X[0], ix0=X[1];
      ATL_INT i;
      prefY(X+24+24-1);
      for (i=0; i < M8; i += 16)
      {
         const double ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2],
                      ia1_0=A0[i+3], ra2_0=A0[i+4], ia2_0=A0[i+5],
                      ra3_0=A0[i+6], ia3_0=A0[i+7], ra4_0=A0[i+8],
                      ia4_0=A0[i+9], ra5_0=A0[i+10], ia5_0=A0[i+11],
                      ra6_0=A0[i+12], ia6_0=A0[i+13], ra7_0=A0[i+14],
                      ia7_0=A0[i+15];
         register double ry0=Y[i+0], iy0=Y[i+1], ry1=Y[i+2], iy1=Y[i+3],
                         ry2=Y[i+4], iy2=Y[i+5], ry3=Y[i+6], iy3=Y[i+7],
                         ry4=Y[i+8], iy4=Y[i+9], ry5=Y[i+10], iy5=Y[i+11],
                         ry6=Y[i+12], iy6=Y[i+13], ry7=Y[i+14], iy7=Y[i+15];

         prefY(Y);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra1_0 * rx0;
         iy1 += ra1_0 * ix0;
         ry2 += ra2_0 * rx0;
         iy2 += ra2_0 * ix0;
         ry3 += ra3_0 * rx0;
         iy3 += ra3_0 * ix0;
         ry4 += ra4_0 * rx0;
         iy4 += ra4_0 * ix0;
         ry5 += ra5_0 * rx0;
         iy5 += ra5_0 * ix0;
         ry6 += ra6_0 * rx0;
         iy6 += ra6_0 * ix0;
         ry7 += ra7_0 * rx0;
         iy7 += ra7_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry1 -= ia1_0 * ix0;
         iy1 += ia1_0 * rx0;
         ry2 -= ia2_0 * ix0;
         iy2 += ia2_0 * rx0;
         ry3 -= ia3_0 * ix0;
         iy3 += ia3_0 * rx0;
         ry4 -= ia4_0 * ix0;
         iy4 += ia4_0 * rx0;
         prefA(A0);
         ry5 -= ia5_0 * ix0;
         iy5 += ia5_0 * rx0;
         ry6 -= ia6_0 * ix0;
         iy6 += ia6_0 * rx0;
         ry7 -= ia7_0 * ix0;
         iy7 += ia7_0 * rx0;


         Y[i+0] = ry0;
         Y[i+1] = iy0;
         Y[i+2] = ry1;
         Y[i+3] = iy1;
         Y[i+4] = ry2;
         Y[i+5] = iy2;
         Y[i+6] = ry3;
         Y[i+7] = iy3;
         Y[i+8] = ry4;
         Y[i+9] = iy4;
         Y[i+10] = ry5;
         Y[i+11] = iy5;
         Y[i+12] = ry6;
         Y[i+13] = iy6;
         Y[i+14] = ry7;
         Y[i+15] = iy7;
      }
      for (i=M8; i < M; i += 2)
      {
         register double ry0 = Y[i], iy0 = Y[i+1];
         const double ra0_0=A0[i], ia0_0=A0[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         Y[i] = ry0;
         Y[i+1] = iy0;
      }
   }
}
